/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-08 * moved clearing of 'hibernated' status from history steps into 'refine_current' - Rolf
03/Mar-07 * removed timer reset on palette cycle speed changes, smoother that way - Rolf
03/Mar-07 * changed memory behaviour to avoid redundant compress/decompress while background crunching - Rolf
03/Mar-05 * added semi-smart max memory adjustment after loading - Rolf
03/Mar-05 * added load and save - Rolf
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

#include <deque>					//needed for Generator.h
#include <vector>					//needed for MaskedImage.h
#include "common.h"
#include "Section.h"				//needed for MaskedImage.h
#include "Span.h"					//needed for MaskedImage.h
#include "downfilter.ah"			//needed for RgbImage.h
#include "x86_timer.ah"				//needed for RgbImage.h
#include "Timing.h"					//needed for RgbImage.h
#include "MaskedImage.h"			//needed for Fractal.h
#include "ExtremeFloat.h"			//needed for Fractal.h
#include "Fractal.h"				//needed for Context.h
#include "Generator.h"				//needed for Context.h
#include "RgbImage.h"				//needed for Context.h
#include "brot_core_interface.h"	//needed for Context.h
#include "HistEntry.h"				//needed for Context.h

#include "Context.h"

//internal stuff
#include <stdio.h>
#include <list>						//needed for Contour.h
#include <vector>					//needed for Contour.h
#include "Boxed.h"
#include "Dumb.h"
#include "Contour.h"

Context::Context(Fractal* first_entry,uint megs):
current(0),
newest(0),
range(256),
sequence(NULL),
decoder(NULL)
{
	//we'll max out at 1 gigabyte and never take more
	if (megs>=(1<<10))	max_memory=1<<30;
	else				max_memory=megs<<20;
	
	reset_all();
	SAFE_DELETE(sequence[current].fractal);
	
	sequence[0]=HistEntry(first_entry,first_entry->get_width(),
		first_entry->get_height(),1,BC_CONTOUR_TRACING);

	decoder->set_size(first_entry->get_width(),first_entry->get_height());
	decoder->set_aa_params(aa_gamma,sequence[current].aa_strength);
	check_mem_usage();
}

Context::~Context()
{
	for (uint i=0;i<range;++i)
	{
		SAFE_DELETE(sequence[i].fractal);
		SAFE_DELETE(sequence[i].generator);
	}
	delete[] sequence;
	SAFE_DELETE(decoder);
}

void
Context::reset_all()
{
    if (sequence)
    {
    	for (uint i=0;i<range;++i)
    	{
    		SAFE_DELETE(sequence[i].fractal);
    		SAFE_DELETE(sequence[i].generator);
    	}
    	delete[] sequence;
    	sequence=NULL;
    }

	SAFE_DELETE(decoder);

	current=0;
	newest=0;
	total_memory=0;
	aa_gamma=1.75f;
	preview_enabled=true;

	sequence=new HistEntry[range];
	
	sequence[0]=HistEntry(new Fractal(100,100),100,100,1,BC_CONTOUR_TRACING);
	total_memory+=sequence[current].fractal->get_used_memory_size();
	decoder=new RgbImage(sequence[current].width,sequence[current].height);
	decoder->set_aa_params(aa_gamma,sequence[current].aa_strength);

    palette_cycle_speed=0.0f;
    cycle_time_remainder=0.0f;
    palette_cycle_timer.reset();
}

bool
Context::step_forward()
{
	//don't step beyond the most recent entry
	if (current==newest) return(false);
	//step forward
	++current;
	sequence[current].fractal->ready();
	sequence[current].hibernated=false;
	decoder->set_size(sequence[current].width,sequence[current].height);
	decoder->set_aa_params(aa_gamma,sequence[current].aa_strength);
	decoder->force_decode();
	check_mem_usage();
	return(true);
}

bool
Context::step_back()
{
	//can't go back from the first entry
	if (current==0) return(false);
	//step backward
	--current;
	decoder->set_size(sequence[current].width,sequence[current].height);
	decoder->set_aa_params(aa_gamma,sequence[current].aa_strength);
	decoder->force_decode();
	check_mem_usage();
	return(true);
}

bool
Context::refine_current()
{
    AbsTimer t;
	if (sequence[current].finished) return(false);

	if (NULL==sequence[current].generator) make_generator(current);
	
	sequence[current].hibernated=false;
	if (!sequence[current].generator->refine_image(0.25f))
	{
		sequence[current].finished=true;
  		decoder->force_decode();
	}
	else
	{
 		decoder->notify_image_change();
	}
	sequence[current].cleverness=sequence[current].fractal->get_cleverness();
	sequence[current].progress=sequence[current].generator->get_progress();
	sequence[current].total_time+=float(t.elapsed_seconds());
	if (sequence[current].finished) SAFE_DELETE(sequence[current].generator);
	return(true);
}

bool
Context::refine(int entry)
{
    AbsTimer t;
	if (sequence[entry].finished)
 	{
 		sequence[entry].fractal->compress();
  		return(false);
	}
	if (NULL==sequence[entry].generator) make_generator(entry);

	if (!sequence[entry].generator->refine_image(0.25f))
	{
		sequence[entry].finished=true;
	}
	sequence[entry].cleverness=sequence[entry].fractal->get_cleverness();
	sequence[entry].progress=sequence[entry].generator->get_progress();
	sequence[entry].total_time+=float(t.elapsed_seconds());
	if (sequence[entry].finished) SAFE_DELETE(sequence[entry].generator);
	return(true);
}

bool
Context::refine_others()
{
	int center=current;
	int step=1;
	int pick=-1;
	while (step<=newest)
	{
		if (((center+step)<=newest)&&((center+step)>=0))
		{
			//find unfinished, non-hibernated fractal
			if ((!sequence[center+step].finished)&&
                (!sequence[center+step].hibernated))
			{
				//select the first one we find
				if (pick==-1)	pick=center+step;
				//shave off all other generators, we only want one
				else SAFE_DELETE(sequence[center+step].generator);
			}
		}
		step=-step;
		if (step>0) ++step;
	}

	if (pick!=-1)
	{
		refine(pick);
		return(true);
	}
	else
	{
		return(false);
	}
}

int
Context::get_generator_type()const
{
	return(sequence[current].generator_type);
}

int
Context::change_generator_type()const
{
	return(set_generator_type(get_generator_type()+1));
}

int
Context::set_generator_type(int type)const
{
	if ((type<0)||(type>BC_MAX_GENERATOR_TYPE)) type=BC_CONTOUR_TRACING;
	if (type!=sequence[current].generator_type)
	{
		sequence[current].generator_type=type;
		SAFE_DELETE(sequence[current].generator);
		sequence[current].finished=false;
	}
	return(type);
}

void
Context::make_generator(int entry)
{
	switch (sequence[entry].generator_type)
	{
	default:
	case (BC_CONTOUR_TRACING):
		sequence[entry].generator=new ContourTracingGenerator(sequence[entry].fractal);
		break;
	case (BC_BOX_MATCHING):
		sequence[entry].generator=new BoxedGenerator(sequence[entry].fractal);
		break;
	case (BC_BRUTE_FORCE):
		sequence[entry].generator=new DumbGenerator(sequence[entry].fractal);
		break;
	}
	sequence[entry].generator->set_preview_quality(sequence[entry].aa_strength*12);
	if (entry==current)
	{
		sequence[entry].generator->set_preview_status(preview_enabled&&
			(sequence[entry].progress==0.0f));
	}
	else
	{
		sequence[entry].generator->set_preview_status(false);
	}
}

void
Context::set_image_properties(int x_res,int y_res,int aa_strength,float aa_gamma)
{
	bool res_changed=false;
	bool aa_changed=false;

	HistEntry& c=sequence[current];

	if ((aa_strength>0)&&(aa_strength<9)&&(aa_strength!=c.aa_strength))
	{
		c.aa_strength=aa_strength;
		aa_changed=true;
	}
	if ((x_res>0)&&(x_res!=c.width))	{ c.width=x_res;	res_changed=true; }
	if ((y_res>0)&&(y_res!=c.height))	{ c.height=y_res;	res_changed=true; }

	if (res_changed||aa_changed)
	{
		c.fractal->set_size(c.aa_strength*c.width,c.aa_strength*c.height);
		c.reset_stats();
		SAFE_DELETE(c.generator);
	}

	if (aa_gamma>0.0f) this->aa_gamma=aa_gamma;
	
	decoder->set_size(c.width,c.height);
	decoder->set_aa_params(this->aa_gamma,c.aa_strength);

	check_mem_usage();
}

RgbImage*
Context::get_decoder()
{
	return(decoder);
}

Fractal*
Context::switch_view()
{
	duplicate_current();
	return(sequence[current].fractal);
}

void
Context::duplicate_current()
{
	//kill generator
	SAFE_DELETE(sequence[current].generator);
	++current;
	for (uint i=current;i<range;++i)
	{
		SAFE_DELETE(sequence[i].fractal);	
		SAFE_DELETE(sequence[i].generator);
	}
	if (current==range)
	{
		--current;
		SAFE_DELETE(sequence[0].fractal);
		SAFE_DELETE(sequence[0].generator);
		for (uint i=0;i<current;++i)
		{
			sequence[i]=sequence[i+1];
		}
	}
	//shallow copy
	sequence[current]=sequence[current-1];
//copy/destroy objects pointed to
	//this copy is intentionally 'the wrong way' to reduce heap fragmentation	
	sequence[current-1].fractal=new Fractal(*(sequence[current].fractal));
	sequence[current].reset_stats();

	newest=current;
	decoder->force_decode();

	check_mem_usage();
}

bool
Context::toggle_preview()
{
	preview_enabled=!preview_enabled;
	if (sequence[current].generator)
		sequence[current].generator->set_preview_status(preview_enabled);
	return(preview_enabled);
}

uint
Context::adjust_max_iterations(float factor)
{
	if ((factor<=0.0f)||(factor==1.0f)) return(sequence[current].fractal->get_max_iterations());
	uint it=sequence[current].fractal->get_max_iterations();
	sequence[current].fractal->set_max_iterations(uint(it*factor));
	if (sequence[current].generator) sequence[current].generator->reset();
	sequence[current].finished=false;
	return(sequence[current].fractal->get_max_iterations());
}

void
Context::reset_current_view()
{
	sequence[current].fractal->reset();
	SAFE_DELETE(sequence[current].generator);
	sequence[current].reset_stats();
	decoder->force_decode();
}

void
Context::zoom(int x,int y,float factor)
{
	duplicate_current();
	HistEntry& c=sequence[current];
	c.fractal->zoom(c.aa_strength*x,c.aa_strength*y,factor);
	decoder->force_decode();
}

void
Context::zoom(float factor)
{
	duplicate_current();
	sequence[current].fractal->zoom(factor);
	decoder->force_decode();
}

void
Context::move(float delta_x,float delta_y)
{
	duplicate_current();
	HistEntry& c=sequence[current];
	int x_steps=int(delta_x*c.width*c.aa_strength);
	int y_steps=int(delta_y*c.height*c.aa_strength);

	if ((x_steps==0)&&(y_steps==0)) return;
	sequence[current].fractal->move(x_steps,y_steps);
	//dirty hack to avoid redoing the preview
	sequence[current].progress=-1.0f;
	decoder->force_decode();
}

const ubyte*
Context::get_indexed_image()const
{
	sequence[current].hibernated=false;
	return(sequence[current].fractal->get_image()->get_data());
}

const ubyte*
Context::get_rgb_image()
{
 	sequence[current].hibernated=false;
   if (palette_cycle_speed!=0.0f)
    {
        cycle_time_remainder+=palette_cycle_timer.elapsed_seconds();
        int cycles=int(cycle_time_remainder*palette_cycle_speed);
        decoder->cycle_palette(cycles);
        cycle_time_remainder-=cycles/palette_cycle_speed;
    }
	decoder->convert_with_palette(sequence[current].fractal->get_image()->get_data());
	return(decoder->get_rgb_data());
}

void
Context::set_palette_cycling_speed(float speed)
{
    palette_cycle_speed=speed;
}

void
Context::check_mem_usage()
{
	//add up memory consumed by fractals
	total_memory=0;
	for (int i=0;i<=newest;++i)
	{
		//don't count the current view
		if (i!=current)	total_memory+=sequence[i].fractal->get_used_memory_size();
	}

	//first reduction pass
	//compress non-hibernated, idle fractals
	int first=current;
	int step=newest;
	bool force_compress=false;			//we'll keep the first one we find
	//overdo compression
	while ((total_memory>((max_memory>>1)+(max_memory>>2)))&&(step!=0))
	{
		if (((first+step)<=newest)&&(first+step>=0))
		{
			if (!sequence[first+step].hibernated)
			{
				if (sequence[first+step].generator)
				{
					if (force_compress)
					{
						SAFE_DELETE(sequence[first+step].generator);
						total_memory-=sequence[first+step].fractal->get_used_memory_size();
						sequence[first+step].fractal->compress();
						total_memory+=sequence[first+step].fractal->get_used_memory_size();
					}
					//spare this one, but the next one will have to go
					else				force_compress=true;
				}
				else
				{
					total_memory-=sequence[first+step].fractal->get_used_memory_size();
					sequence[first+step].fractal->compress();
					total_memory+=sequence[first+step].fractal->get_used_memory_size();
				}
			}
		}
		step=-step;
		if (step>0) --step;
	}
	//second reduction pass
	//hibernate fractals until we fit the memory constraint
	first=current;
	step=newest;

	while ((total_memory>max_memory)&&(step!=0))
	{
		if (((first+step)<=newest)&&(first+step>=0))
		{
			if (!sequence[first+step].hibernated)
			{
				total_memory-=sequence[first+step].fractal->get_used_memory_size();
				sequence[first+step].fractal->hibernate();
				sequence[first+step].hibernated=true;
				sequence[first+step].reset_stats();
				total_memory+=sequence[first+step].fractal->get_used_memory_size();
			}
		}
		step=-step;
		if (step>0) --step;
	}
}


bool
Context::save(const char* filename,bool everything)
{
	FILE* of=fopen(filename,"wb");
	if (NULL==of) return(false);
//write signature/version
	fputs("brot_core\n",of);
	fputs("Context\n",of);
	fputs("v000\n",of);
//write significant context data
	fwrite(&max_memory,sizeof(uint),1,of);
	fwrite(&range,sizeof(int),1,of);
	fwrite(&current,sizeof(int),1,of);
	fwrite(&newest,sizeof(int),1,of);

	fwrite(&preview_enabled,sizeof(bool),1,of);
	fwrite(&aa_gamma,sizeof(float),1,of);
	fwrite(&palette_cycle_speed,sizeof(float),1,of);
//we don't need the decoder, but we sure can use the current palette
	decoder->save_palette(of);

//step through all history entries and save them
	for (int i=0;i<=newest;++i)
	{
		fputs("HistEntry\n",of);
		sequence[i].save(of,everything);
	}

	fclose(of);
	return(true);
}


bool
Context::load(const char* filename)
{
	FILE* inf=fopen(filename,"rb");
	if (NULL==inf) return(false);
//read signature/version
	char signature[16];
	fgets(signature,16,inf);
	if (0!=strcmp(signature,"brot_core\n")) return(false);
	fgets(signature,16,inf);
	if (0!=strcmp(signature,"Context\n")) return(false);
	char version[16];
	fgets(version,16,inf);
	if (0!=strcmp(version,"v000\n")) return(false);

//destroy all controlled objects, except for the decoder
	for (int nuke=0;nuke<=newest;++nuke)
	{
		SAFE_DELETE(sequence[nuke].fractal);
		SAFE_DELETE(sequence[nuke].generator);
	}
	delete[] sequence;

//read context stuff
	uint desired_memory;

	fread(&desired_memory,sizeof(uint),1,inf);
	fread(&range,sizeof(int),1,inf);
	fread(&current,sizeof(int),1,inf);
	fread(&newest,sizeof(int),1,inf);

	fread(&preview_enabled,sizeof(bool),1,inf);
	fread(&aa_gamma,sizeof(float),1,inf);
	fread(&palette_cycle_speed,sizeof(float),1,inf);
//restore palette
	ubyte palette[768];
	fread(palette,sizeof(ubyte),768,inf);
	decoder->set_palette(palette);
//restore history entries
	sequence=new HistEntry[range];

	int i;
	for (i=0;i<=newest;++i)
	{
		fgets(signature,16,inf);
		if (0!=strcmp(signature,"HistEntry\n"))
		{
			fclose(inf);
			return(false);
		}
		sequence[i]=HistEntry(inf);
	}
	fclose(inf);

	decoder->set_size(sequence[current].width,sequence[current].height);
	decoder->set_aa_params(aa_gamma,sequence[current].aa_strength);

	cycle_time_remainder=0.0f;
	palette_cycle_timer.reset();


	//add up memory consumed by fractals
	total_memory=0;
	for (i=0;i<=newest;++i)
	{
		//don't count the current view
		if (i!=current)	total_memory+=sequence[i].fractal->get_used_memory_size();
	}

	//don't go bonkers if the loaded data doesn't justify its mem requirements
	if ((total_memory>(desired_memory>>2))&&
		(desired_memory>max_memory))
	{
		max_memory=desired_memory;
	}

	return(true);
}
