/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

//stuff for understanding headers
#include <vector>			//needed for MaskedImage.h
#include "common.h"
#include "ExtremeFloat.h"	//needed for Fractal.h
#include "Section.h"		//needed for MaskedImage.h
#include "Span.h"			//needed for MaskedImage.h
#include "MaskedImage.h"	//needed for Fractal.h
#include "Fractal.h"		//needed for Generator.h
#include <deque>			//needed for Generator.h
#include "Generator.h"   	//needed for Contour.h
#include "x86_timer.ah"		//needed for Timing.h"
#include "Timing.h"			//needed for Contour.h
#include <list>				//needed for Contour.h
#include "Contour.h"

//internal stuff
#include <algorithm>

ContourTracingGenerator::ContourTracingGenerator(Fractal* fractal):
Generator(fractal)
{
	reset();
}

ContourTracingGenerator::~ContourTracingGenerator()
{
}

bool
ContourTracingGenerator::refine_image(float max_time)
{
	fractal->ready();
	AbsTimer refine_timer;

	bool ret=preview(max_time);

	while (refine_timer.elapsed_seconds()<max_time)
	{
		if (unfinished_contour)
		{
			ret=true;
			trace(refine_timer);
			if (!unfinished_contour) fill_traced_contour();
			continue;
		}
		if (spans.empty()) return(ret);

		ret=true;
		
  		//grab a span that's not finished yet
    	Span* sp=&spans.front();
		
		//get reference color from the start of the span
		uint reference=(*fractal)(sp->x0,sp->y);
		cx=x0=sp->x0;
		cy=y0=sp->y;

		//discard single pixel spans
		if (sp->x0>=sp->x1)
		{
			spans.pop_front();
   			continue;
		}
		else
		{
			//otherwise shorten it by one pixel
			++sp->x0;
		}

		//if no neighbor pixel are equal, we don't need to trace
		if ((reference!=(*fractal)(x0+1,y0))&&
			(reference!=(*fractal)(x0-1,y0))&&
			(reference!=(*fractal)(x0,y0+1))&&
			(reference!=(*fractal)(x0,y0-1)))
		{
			continue;
		}
		else
		{
			unfinished_contour=true;
			color=reference;
			dx=-1;	dy=-1;
			//find contour
			trace(refine_timer);
			if (!unfinished_contour) fill_traced_contour();
		}
	}
	return(ret);
}

void
ContourTracingGenerator::trace(AbsTimer& timer)
{
	while (timer.elapsed_seconds()<0.6)
	{
	//trace horizontal edges
		while (1)
		{
			if (color==(*fractal)(cx+dx,cy)) cx+=dx;
			else
			{
				contour.push_back(ContourPoint(cx,cy));
				dy=-dy;
				break;
			}
			if ((cx==x0)&&(cy==y0))
			{
				unfinished_contour=false;
				return;
			}
			if (color==(*fractal)(cx,cy+dy))
			{
				dx=-dx;
				break;
			}
		}
	//trace vertical edges
		while (1)
		{
			if (color==(*fractal)(cx,cy+dy)) cy+=dy;
			else { dx=-dx; break; }
			if ((cx==x0)&&(cy==y0))
			{
				contour.push_back(ContourPoint(cx,cy));
				unfinished_contour=false;
				return;
			}
			if (color==(*fractal)(cx+dx,cy)) { dy=-dy; break; }
			contour.push_back(ContourPoint(cx,cy));
		}
	}
}

void
ContourTracingGenerator::reset()
{
	width=fractal->get_width();
	height=fractal->get_height();

	spans.clear();
	contour.clear();

	for (uint y=0;y<height;++y)
	{
  		spans.push_back(Span(0,width-1,y));
	}
	reset_preview();
	unfinished_contour=false;
	y0=0;
}

void
ContourTracingGenerator::fill_traced_contour()//const int value)
{
	if (contour.empty()) return;

	//sort contour points
	std::sort(contour.begin(),contour.end());
	//build a sorted list of spans to fill
	std::vector <Span> fill_spans;
	int cp=0;
	while (cp<contour.size())
	{
		ContourPoint p0=contour[cp];
		++cp;
		ContourPoint p1;

		if (cp==contour.size()) p1=p0;
		else
		{
			p1=contour[cp];
			if (p1.y==p0.y) ++cp;
			else			p1=p0;
		}
		fill_spans.push_back(Span(p0.x,p1.x,p0.y));
	}

	fractal->get_image()->fill(fill_spans,color);
	//remove the filled spans, they're done
	subtract_spans(fill_spans);
	contour.clear();
}

void
ContourTracingGenerator::subtract_spans(const std::vector <Span>& to_remove)
{
    if (spans.empty()) return;
	//target spans we operate on and change
	std::list<Span>::iterator t=spans.begin();
	int s=0;		//source (to_remove[s])

	//this one is just for kicks but still saves some overhead
	const int s_max=to_remove.size();

 	//the left part will be appended to spans[], as it won't be needed more than once
	while ((t!=spans.end())&&(s<s_max))
	{
  		if (!t->overlaps(to_remove[s]))
		{
			if (to_remove[s]<*t)	++s;
			else					++t;
			continue;
		}
		//spans overlap, we need to check _how_ and cut target_span
 		const Span& rem_span=to_remove[s];

		if (rem_span.x0>t->x0)
		{
            //rem_span starts a bit more to the right
            //cut off the left part of target_span and try again
            Span left_part=*t;
            left_part.x1=rem_span.x0-1;
            spans.insert(t,left_part);
            t->x0=rem_span.x0;
        }
		
		if ((t->x0>=rem_span.x0)&&(t->x1<=rem_span.x1))
		{
            //rem_span completely 'occludes' target_span. KILL IT!!!!!111
			t=spans.erase(t);
			continue;
        }

        if (rem_span.x1<t->x1)
        {
            //rem_span is now guaranteed to have the same x0 as target_span
            //but it's a little shorter
            //reduce target_span to it's right part, beyond rem_span
            t->x0=rem_span.x1+1;
			//rem_spam _cannot_ overlap the current span anymore
			++s;
        }
    }
}

float
ContourTracingGenerator::get_progress()
{
	if (height!=0)	return(float(y0)/height);
	else			return(0.0f);
}
