/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

#include "common.h"

#include "ExtremeFloat.h"

//internal stuff
#include "xfloat.ah"
#include <memory.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>				//strchr
#include <stdlib.h>				//atoi, atof


ExtremeFloat::ExtremeFloat()
{
	set_to_zero();
}

void
ExtremeFloat::set_to_zero()
{
	mantissa[0]=0;
	mantissa[1]=0;
	mantissa[2]=0;
	mantissa[3]=0;
	mantissa[4]=0;
	mantissa[5]=0;
	exponent=0;
	sign=0;
}

ExtremeFloat::ExtremeFloat(double a)
{
	x86_double_to_xfloat(this,&a);
}

ExtremeFloat::ExtremeFloat(const int a)
{
	x86_int_to_xfloat(this,a);
}

ExtremeFloat::ExtremeFloat(const uint a)
{
	x86_int_to_xfloat(this,a);
}

ExtremeFloat::ExtremeFloat(const char*const c_string)
{
	const char* exponent_start=strchr(c_string,'x');
	if (NULL==exponent_start)
	{
		double a=atof(c_string);
		x86_double_to_xfloat(this,&a);
	}
	else
	{
		set_to_zero();
		exponent=atoi(exponent_start+1);
		const char* digit=c_string;
		//skip leading whitespace
		while ((*digit==' ')||(*digit=='\t')) ++digit;
		if (*digit=='-')	{	sign=1;	++digit; }
		else
		if (*digit=='+')	{	sign=0;	++digit; }

		int significance=8*6-1;
		while (significance>=0)
		{
			char c=*digit;	++digit;
			if (c==0) break;
			if (c=='.')
			{
				exponent-=4*(significance-(8*6-2));
				continue;
			}
			if ((c>='0')&&(c<='9'))
			{
				mantissa[significance>>3] |= (c-'0')<<(4*(significance&7));
			}
			else
			{
				c&=~0x20;
				if ((c>='A')&&(c<='F'))
				{
					mantissa[significance>>3] |= (c-'A'+10)<<(4*(significance&7));
				}
				else
					break;
			}
			--significance;
		}
	}
}


double
ExtremeFloat::to_double() const
{
	return(x86_xfloat_to_double(this));
}

ExtremeFloat
ExtremeFloat::operator * (const ExtremeFloat& rhs) const
{
	if (this->is_zero()||rhs.is_zero())
	{
		ExtremeFloat ret;
		return(ret);
	}
	else
	{
		ExtremeFloat tmp(*this);
		tmp*=rhs;
		return(tmp);
	}
}

ExtremeFloat
ExtremeFloat::operator + (const ExtremeFloat& rhs) const
{
	ExtremeFloat tmp(*this);
	tmp+=rhs;
	return(tmp);
}

ExtremeFloat
ExtremeFloat::operator - (const ExtremeFloat& rhs) const
{
	ExtremeFloat tmp(*this);
	tmp-=rhs;
	return(tmp);
}

void
ExtremeFloat::operator += (const ExtremeFloat& rhs)
{
	if (rhs.is_zero()) return;
	if (this->is_zero())
	{
		*this=rhs;
		return;
	}

	uint order=(2*x86_xfloat_abs_order(this,&rhs))^sign^rhs.sign;
	switch (order)
	{
	default:
	case(0):	//same sign, equal magnitude
		++exponent;
		break;
	case(1):	//different signs, equal magnitude
		set_to_zero();
		break;
	case(2):	//same sign, abs(*this)>abs(rhs)
		{
			ExtremeFloat tmp(rhs);
			x86_xfloat_abs_add(&tmp,this);
			*this=tmp;
		}
		break;
	case(3):	//different signs, abs(*this)>abs(rhs)
		{
			ExtremeFloat tmp(rhs);
			x86_xfloat_abs_subr(&tmp,this);
			*this=tmp;
			sign^=1;
		}
		break;
	case(4):	//same sign, abs(rhs)>abs(*this)
		{
			x86_xfloat_abs_add(this,&rhs);
		}
		break;
	case(5):	//different signs, abs(rhs)>abs(*this)
		x86_xfloat_abs_subr(this,&rhs);
		sign^=1;
		break;
	}
}
void
ExtremeFloat::operator -= (const ExtremeFloat& rhs)
{
	if (rhs.is_zero()) return;
	if (this->is_zero())
	{
		*this=rhs;
		sign^=1;
		return;
	}

	uint order=(2*x86_xfloat_abs_order(this,&rhs))^sign^rhs.sign;
	switch (order)
	{
	default:
	case(0):	//same sign, equal magnitude
		set_to_zero();
		break;
	case(1):	//different signs, equal magnitude
		++exponent;
		break;
	case(2):	//same sign, abs(*this)>abs(rhs)
		{
			ExtremeFloat tmp(*this);
			*this=rhs;
			x86_xfloat_abs_subr(this,&tmp);
		}
		break;
	case(3):	//different signs, abs(*this)>abs(rhs)
		{
			ExtremeFloat tmp(rhs);
			tmp.sign^=1;
			x86_xfloat_abs_add(&tmp,this);
			*this=tmp;
		}
		break;
	case(4):	//same sign, abs(rhs)>abs(*this)
		x86_xfloat_abs_subr(this,&rhs);
		sign^=1;
		break;
	case(5):	//different signs, abs(rhs)>abs(*this)
		x86_xfloat_abs_add(this,&rhs);
		break;
	}
}

void
ExtremeFloat::operator *= (const ExtremeFloat& rhs)
{
	if (&rhs==this)	x86_xfloat_square(this);
	else
	{
		if ((this->is_zero())||(rhs.is_zero())) set_to_zero();
		else	x86_xfloat_mul(this,&rhs);
	}
}

ExtremeFloat
ExtremeFloat::operator * (const uint rhs) const
{
	ExtremeFloat ret(*this);
	ret*=rhs;
	return(ret);
}

ExtremeFloat
ExtremeFloat::operator * (const int rhs) const
{
	ExtremeFloat ret(*this);
	ret*=rhs;
	return(ret);
}

void
ExtremeFloat::operator *= (const int rhs)
{
	if (rhs<0)
	{
		*this*=(uint)(-rhs);
		sign^=1;
	}
	else
	{
		*this*=(uint)(rhs);
	}
}

void
ExtremeFloat::operator *= (const uint rhs)
{
	if ((0==rhs)||this->is_zero())
	{
		set_to_zero();
		return;
	}
	else
	{
		x86_xfloat_mul_uint(this,rhs);
	}
}

//************************************************************************


bool
ExtremeFloat::operator >= (double const& rhs) const
{
	ExtremeFloat xrhs(rhs);
	if (sign!=xrhs.sign) return(xrhs.sign!=0);

	uint order=x86_xfloat_abs_order(this,&xrhs)>>sign;
	return(order==1);
}

void
ExtremeFloat::to_string(char* target_buffer) const
{
	if (is_zero()) strcpy(target_buffer,"0");

	if (sign)	sprintf(target_buffer,"-%1X.",mantissa[5]>>28);
	else		sprintf(target_buffer,"+%1X.",mantissa[5]>>28);
	target_buffer+=3;

	sprintf(target_buffer,"%07X%08X%08X%08X%08X%08X",
		mantissa[5]&0xFFFFFFF,mantissa[4],mantissa[3],mantissa[2],mantissa[1],mantissa[0]);
	char* end=target_buffer+46;
	while ((end>(target_buffer+1))&&(*end=='0')) --end;
	sprintf(end+1,"x%+d",exponent);
}
