/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-08 * fixed crash on 'zoom too early' - Rolf
03/Mar-05 * added proper LGPL notice - Rolf
03/Mar-03 * made move function actually do _something_ clever - Rolf
          * started changelog
*/


#include <memory.h>
#include <stdlib.h>
#include <stdio.h>
#include <vector>			//needed for MaskedImage.h
#include "common.h"
#include "Span.h"			//needed for MaskedImage.h
#include "Section.h"		//needed for MaskedImage.h
#include "MaskedImage.h"
#include "ExtremeFloat.h"
#include "Fractal.h"

Fractal::Fractal(int width,int height):
image(NULL),
column_to_real_extreme(NULL),
column_to_real_double(NULL),
line_to_imaginary_extreme(NULL),
line_to_imaginary_double(NULL),
max_it(150),
width(0),height(0),
center_r(0.0),center_i(0.0),radius(2.0)
{
	set_size(width,height);
}

Fractal::Fractal(FILE* inf):
image(NULL),
column_to_real_extreme(NULL),
column_to_real_double(NULL),
line_to_imaginary_extreme(NULL),
line_to_imaginary_double(NULL)
{
	char signature[16];
	fgets(signature,16,inf);
	if (0!=strcmp(signature,"v000\n"))
	{
//FIXME
	}
	else
	{
		fread(&center_r,sizeof(ExtremeFloat),1,inf);
		fread(&center_i,sizeof(ExtremeFloat),1,inf);
		fread(&r0,sizeof(ExtremeFloat),1,inf);
		fread(&i0,sizeof(ExtremeFloat),1,inf);
		fread(&r_step,sizeof(double),1,inf);
		fread(&i_step,sizeof(double),1,inf);
		fread(&radius,sizeof(double),1,inf);
		fread(&width,sizeof(int),1,inf);
		fread(&height,sizeof(int),1,inf);
		fread(&max_it,sizeof(uint),1,inf);
		fread(&iterated_samples,sizeof(uint),1,inf);
		total_samples=width*height;

		char signature[16];
		fgets(signature,16,inf);
		if (0==strcmp("MaskedImage\n",signature))
		{
			image=new MaskedImage(inf);
		}
		ready(false);			//prepare derived data, but don't decompress the image just yet
	}
}


Fractal::~Fractal()
{
	SAFE_FREE(line_to_imaginary_double);
	SAFE_FREE(column_to_real_double);
	SAFE_FREE(line_to_imaginary_extreme);
	SAFE_FREE(column_to_real_extreme);
	SAFE_DELETE(image);
}

bool
Fractal::save(FILE* of,bool everything)
{
	fputs("v000\n",of);

	fwrite(&center_r,sizeof(ExtremeFloat),1,of);
	fwrite(&center_i,sizeof(ExtremeFloat),1,of);
	fwrite(&r0,sizeof(ExtremeFloat),1,of);
	fwrite(&i0,sizeof(ExtremeFloat),1,of);
	fwrite(&r_step,sizeof(double),1,of);
	fwrite(&i_step,sizeof(double),1,of);
	fwrite(&radius,sizeof(double),1,of);
	fwrite(&width,sizeof(int),1,of);
	fwrite(&height,sizeof(int),1,of);
	fwrite(&max_it,sizeof(uint),1,of);

	if (image&&everything)
	{
		fwrite(&iterated_samples,sizeof(uint),1,of);
		fputs("MaskedImage\n",of);
		return(image->save(of));
	}
	else
	{
  		uint spam=0;
		fwrite(&spam,sizeof(uint),1,of);
		fputs("NULL\n",of);
		return(true);
	}
}

int
Fractal::operator()(const int x,const int y)
{
	if ((x<0)||(x>=width)||(y<0)||(y>=height)) return(-1);

	if (!image->check_mask(x,y))
	{
		++iterated_samples;
		uint it;
		if (extreme_precision)
		{
			it=iterate_mandelbrot_extreme(	column_to_real_extreme[x],
											line_to_imaginary_extreme[y]);
		}
		else
		{
			it=iterate_mandelbrot_double(	column_to_real_double[x],
											line_to_imaginary_double[y]);
		}

		if (it==max_it)
		{
			image->put_and_mask(x,y,0);
			return(0);
		}
		else
		{
			int val=((it-1)%255)+1;
			image->put_and_mask(x,y,val);
			return(val);
		}

	}
	return(image->get(x,y));
}

void
Fractal::set_params(const ExtremeFloat& center_r,const ExtremeFloat& center_i,const double radius)
{
	if (radius>=2.0)
	{
		this->center_r=0.0;
		this->center_i=0.0;
		this->radius=2.0;
	}
	else
	{
		this->center_r=center_r;
		this->center_i=center_i;
		this->radius=radius;
	}
	
	fit_params_to_size();
 	reset();
}

void
Fractal::fit_params_to_size()
{
	SAFE_FREE(line_to_imaginary_double);
	SAFE_FREE(column_to_real_double);
	SAFE_FREE(line_to_imaginary_extreme);
	SAFE_FREE(column_to_real_extreme);

	double dr,di;
	if (width>=height)
	{
		r0=center_r-radius;
		dr=radius*2;
		double aspect=height*(1.0/width);
		i0=center_i+aspect*radius;
		di=-dr*aspect;
	}
	else
	{
		i0=center_i+radius;
		di=-radius*2;
		double aspect=width*(1.0/height);
		r0=center_r-radius*aspect;
		dr=-di*aspect;
	}
	r_step=dr/width;
	i_step=di/height;
}

void
Fractal::set_size(int width,int height)
{
	if ((width<=0)||(height<=0)) return;
	if ((width==this->width)&&(height==this->height))return;
	this->width=width;
	this->height=height;

	iterated_samples=0;
	total_samples=width*height;
	fit_params_to_size();
	reset();
}

void
Fractal::set_max_iterations(uint max_it)
{
	//due to explicit loop optimization, we can't accept anything less than two
	if (max_it<2) max_it=2;

	if (image)
 	{
		image->decompress();
		//iteration depth has incresed
		//we can reuse non-black samples, because they finished before reaching this limit
		if (max_it>this->max_it)
		{
			//reset cache _only_ for black samples
			for (uint y=0;y<height;++y)
			{
				for (uint x=0;x<width;++x)
				{
					if (image->get(x,y)==0) image->clear_mask(x,y);
				}
			}
		}
		else
			image->clear_mask();
	}
	this->max_it=max_it;
	iterated_samples=0;
}

void
Fractal::ready(bool decompress)
{
	if (decompress)
	{
		if (NULL==image)
		{
			image=new MaskedImage(width,height);
		}
		else image->decompress();
	}

	if (NULL==column_to_real_extreme)
	{
		SAFE_FREE(line_to_imaginary_double);
		SAFE_FREE(column_to_real_double);
		SAFE_FREE(line_to_imaginary_extreme);
		column_to_real_extreme=(ExtremeFloat*)calloc(width,sizeof(ExtremeFloat));
		line_to_imaginary_extreme=(ExtremeFloat*)calloc(height,sizeof(ExtremeFloat));
		column_to_real_double=(double*)calloc(width,sizeof(double));
		line_to_imaginary_double=(double*)calloc(height,sizeof(double));

 		uint x,y;
 		for (x=0;x<width;++x)
			column_to_real_extreme[x]=r0+r_step*x;

		for (y=0;y<height;++y)
			line_to_imaginary_extreme[y]=i0+i_step*y;

		try_with_doubles();
	}
}

void
Fractal::try_with_doubles()
{
	uint x,y;
    //convert grid coordinates from ExtremeFloats to doubles
	for (x=0;x<width;++x)
		column_to_real_double[x]=column_to_real_extreme[x].to_double();

	for (y=0;y<height;++y)
		line_to_imaginary_double[y]=line_to_imaginary_extreme[y].to_double();

    //check if any neighbor doubles are equal, if yes, assume that ExtremeFloats are needed
	extreme_precision=false;
	double last_r=column_to_real_double[0];
	for (x=1;x<width;++x)
	{
		if (column_to_real_double[x]==last_r)
		{
			extreme_precision=true;
			break;
		}
		last_r=column_to_real_double[x];
	}
	double last_i=line_to_imaginary_double[0];
	for (y=1;y<height;++y)
	{
		if (line_to_imaginary_double[y]==last_i)
		{
			extreme_precision=true;
			break;
		}
		last_i=line_to_imaginary_double[y];
	}
}

void
Fractal::reset()
{
	SAFE_DELETE(image);
	iterated_samples=0;
}

void
Fractal::zoom(const int x,const int y,const float factor)
{
	ExtremeFloat new_r;
	ExtremeFloat new_i;
	if ((x>=0)&&(x<width)&&(column_to_real_extreme!=NULL))	new_r=column_to_real_extreme[x];
	else													new_r=r0+x*r_step;
	if ((y>=0)&&(y<height)&&(line_to_imaginary_extreme!=NULL))	new_i=line_to_imaginary_extreme[y];
	else														new_i=i0+y*i_step;

	set_params(new_r,new_i,radius/factor);
}

void
Fractal::move(int delta_x,int delta_y)
{
    delta_y=-delta_y;
	delta_x&=~7;		//we want delta_x as whole mulitples of 8
    
    if ((delta_x==0)&&(delta_y==0)) return;
    if ((labs(delta_x)>=width)||(labs(delta_y)>=height)||
        (NULL==image)||(NULL==column_to_real_extreme)||(NULL==image))
    {
	   set_params(center_r+delta_x*r_step,center_i+delta_y*i_step,radius);
	   return;
    }

    image->move(delta_x,delta_y);
    
    if (delta_x!=0)
    {
        if (delta_x>0)
        {
            r0=column_to_real_extreme[delta_x];
            memmove(column_to_real_extreme,column_to_real_extreme+delta_x,
                (width-delta_x)*sizeof(ExtremeFloat));

            for (uint x=width-delta_x;x<width;++x)
                column_to_real_extreme[x]=r0+x*r_step;
        }
        else
        {
            r0=column_to_real_extreme[0]+delta_x*r_step;
            memmove(column_to_real_extreme-delta_x,column_to_real_extreme,
                (width+delta_x)*sizeof(ExtremeFloat));

            for (uint x=0;x<(-delta_x);++x)
                column_to_real_extreme[x]=r0+x*r_step;
        }
    }
    if (delta_y!=0)
    {
        if (delta_y>0)
        {
            i0=line_to_imaginary_extreme[delta_y];
            memmove(line_to_imaginary_extreme,line_to_imaginary_extreme+delta_y,
                (height-delta_y)*sizeof(ExtremeFloat));

            for (uint y=height-delta_y;y<height;++y)
                line_to_imaginary_extreme[y]=i0+y*i_step;
        }
        else
        {
            i0=line_to_imaginary_extreme[0]+delta_y*i_step;
            memmove(line_to_imaginary_extreme-delta_y,line_to_imaginary_extreme,
                (height+delta_y)*sizeof(ExtremeFloat));

            for (uint y=0;y<(-delta_y);++y)
                line_to_imaginary_extreme[y]=i0+y*i_step;
        }
    }
    center_r=column_to_real_extreme[width>>1];
    center_i=line_to_imaginary_extreme[height>>1];
    try_with_doubles();
    iterated_samples=0;
}


float
Fractal::get_cleverness()
{
	if (total_samples==0) return(0.0f);
	
	return(1.0f-float(iterated_samples)/total_samples);
}

MaskedImage*
Fractal::get_image()
{
	if (NULL==image) image=new MaskedImage(width,height);
	image->decompress();
	return(image);
}

Fractal::Fractal(const Fractal& rhs):
image(NULL),
column_to_real_extreme(NULL),line_to_imaginary_extreme(NULL),
column_to_real_double(NULL),line_to_imaginary_double(NULL)
{
	//delegate to assignment
	*this=rhs;
}

Fractal&
Fractal::operator =(const Fractal& rhs)
{
	//don't shoot ourselves in the foot
	if (this==&rhs) return(*this);
	//destroy our own controlled data
	SAFE_DELETE(image);
	SAFE_FREE(line_to_imaginary_double);
	SAFE_FREE(column_to_real_double);
	SAFE_FREE(line_to_imaginary_extreme);
	SAFE_FREE(column_to_real_extreme);
 
 	//plain copy of all members
	memcpy(this,&rhs,sizeof(Fractal));
	//all pointers!=NULL need deep copy
	if (rhs.image)
	{
		image=new MaskedImage(*rhs.image);
	}
	if (rhs.column_to_real_extreme)
	{
		column_to_real_extreme=(ExtremeFloat*)malloc(width*sizeof(ExtremeFloat));
		memcpy(column_to_real_extreme,rhs.column_to_real_extreme,width*sizeof(ExtremeFloat));
	}
	if (rhs.line_to_imaginary_extreme)
	{
		line_to_imaginary_extreme=(ExtremeFloat*)malloc(height*sizeof(ExtremeFloat));
		memcpy(line_to_imaginary_extreme,rhs.line_to_imaginary_extreme,height*sizeof(ExtremeFloat));
	}
	if (rhs.column_to_real_double)
	{
		column_to_real_double=(double*)malloc(width*sizeof(double));
		memcpy(column_to_real_double,rhs.column_to_real_double,width*sizeof(double));
	}
	if (rhs.line_to_imaginary_double)
 	{	
 		line_to_imaginary_double=(double*)malloc(height*sizeof(double));
 		memcpy(line_to_imaginary_double,rhs.line_to_imaginary_double,height*sizeof(double));
	}
	return(*this);
}

void
Fractal::hibernate()
{
	SAFE_DELETE(image);
	SAFE_FREE(line_to_imaginary_double);
	SAFE_FREE(column_to_real_double);
	SAFE_FREE(line_to_imaginary_extreme);
	SAFE_FREE(column_to_real_extreme);
	iterated_samples=0;
}

uint
Fractal::get_used_memory_size()const
{
	uint ret=sizeof(Fractal);
	if (image) ret+=image->get_used_memory_size();
	if (line_to_imaginary_extreme) ret+=height*sizeof(ExtremeFloat);
	if (line_to_imaginary_double) ret+=height*sizeof(double);
	if (column_to_real_extreme) ret+=width*sizeof(ExtremeFloat);
	if (column_to_real_double) ret+=width*sizeof(double);
	return(ret);
}

uint
Fractal::estimate_required_memory()const
{
	uint ret=sizeof(Fractal);
	ret+=MaskedImage::estimate_required_memory(width,height);
	ret+=(height+width)*sizeof(ExtremeFloat);	//line/column caches
	ret+=(height+width)*sizeof(double);
	return(ret);
}

uint
Fractal::iterate_mandelbrot_extreme(const ExtremeFloat& cr,const ExtremeFloat& ci)const
{
	ExtremeFloat zr=cr;
	ExtremeFloat zi=ci;

	uint it=1;
	do
	{
		ExtremeFloat i2=zi;	i2*=i2;

		zi*=zr;	zi.scale(1);	zi+=ci;
		zr*=zr;
		if ((zr+i2).above_or_equal_pow2(2)) break;

		zr-=i2; zr+=cr;
		++it;
	} while (it<max_it);

	return(it);
}

uint
Fractal::iterate_mandelbrot_double(const double cr,const double ci)const
{
	double zr=cr;
	double zi=ci;
	
	uint it=1;
	do
	{
		double i2=zi*zi;
		zi=zi*zr*2+ci;
		zr=zr*zr;
		if ((zr+i2)>=4.0) break;

		zr=zr-i2+cr;
		++it;
	} while (it<max_it);

  return(it);
}

