/*
this is part of brot_core.dll, a fractal generation library
Copyright (C) 2003 Rolf Neuberger

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
changes:
03/Mar-05 * added proper LGPL notice, started changelog - Rolf

*/

class
Fractal
{
private:
	Fractal();	//hide default constructor
public:
	Fractal(const int width,const int height);
	Fractal(FILE* inf);
	~Fractal();

	bool save(FILE* of,bool everything);

	//copy constructor
	Fractal(const Fractal& rhs);
	//assignment
	Fractal& operator =(const Fractal& rhs);

	//caching iteration result retrieval
	int operator() (const int x,const int y);
	int sample(const int x,const int y) {return((*this)(x,y));}

	//set fractal parameters
	void set_params(const ExtremeFloat& center_r,const ExtremeFloat& center_i,const double radius);
	//zoom in or out on pixel coordinate x/y
	void zoom(const int x,const int y,const float factor);
	//zoom in or out on current image center
	void zoom(const double factor) {
		set_params(center_r,center_i,
				   radius/factor); }
	//move the center point around without zooming
	void move(int delta_x,int delta_y);
	//set image size in pixels
	void set_size(const int width,const int height);
	//set maximum iteration depth
	void set_max_iterations(const uint max_it);

	//forget all finished samples and restart
	void reset();
	//get ready for processing (image retrieval/iterations)
	//the parameter is mostly useful for loading batches of fractals
	//passing 'false' keeps the image compressed but inits all other derived data
	void ready(bool decompress=true);

//accessors
	//get fractal parameters
	//center/radius representation
	const ExtremeFloat& get_center_r() const { return(center_r); }
	const ExtremeFloat& get_center_i() const { return(center_i); }
	double get_radius() const { return(radius); }
	//get maximum iteration count
	uint get_max_iterations() const { return(max_it); }

	//get pointer to image data (array of iteration depths)
	MaskedImage* get_image();
	//get image dimensions in pixels
	int get_width() const { return(width); }
	int get_height() const { return(height); }
	

//reduce memory consumption.
//clients of the class _must_ call ready() before asking for samples
	//compress to conserve space when not currently in use, lossless
	void compress() {if (image) image->compress(); }
	//remove all controlled memory, keeps around parameters but destroys all progress
	void hibernate();
	//get memory consumed by instance in its current state (ready/compressed/hibernated)
	uint get_used_memory_size()const;
	//get memory usage estimate for ready state (meaningful for hibernated fractals)
	uint estimate_required_memory()const;

//statistics and generator utilities
	float get_cleverness();
private:
	uint iterate_mandelbrot_extreme(const ExtremeFloat& r,const ExtremeFloat& i)const;
	uint iterate_mandelbrot_double(const double r,const double i)const;
	
	//derive origin and step from params and window size/aspect ratio
	void fit_params_to_size();
	//decide whether to use doubles or ExtremeFloats
	void try_with_doubles();

private:
//fractal parameters
	//center
	ExtremeFloat center_r;
	ExtremeFloat center_i;
	//origin (upper left corner), used internally
	ExtremeFloat r0;
	ExtremeFloat i0;
	double r_step;				//step in r for one pixel
	double i_step;				//step in i for one pixel

	double radius;

	//result array plus bitmask, indicating actually calculated samples
	MaskedImage* image;

	//image resolution in samples
	int width;
	int height;

	//cache r0+r_step*x
	ExtremeFloat* column_to_real_extreme;
	double* column_to_real_double;

	//cache i0+i_step*y
	ExtremeFloat* line_to_imaginary_extreme;
	double* line_to_imaginary_double;

	//statistics
	uint total_samples;
	uint iterated_samples;

	//extreme precision required? (if not, we use regular doubles)
	bool extreme_precision;
protected:
	uint max_it;				//maximum iteration depth
};
